package ch.epfl.collections;

import java.util.Arrays;

/**
 * Implementation of the StringList interface based on a array.
 * 
 * @see StringList
 * @see StringArrayListTest
 * @see StringArrayListIterator
 *
 */
public class StringArrayList implements StringList {
    /**
     * Internal storage of the list elements.
     */
    private String[] array = new String[1];

    /**
     * Indicates the current number of filled locations in the
     * internal array storage. 
     */
    private int size = 0;

    @Override
    public boolean isEmpty() {
        return size == 0;
    }

    @Override
    public int size() {
        return size;
    }

    /**
     * Adds a new element to the list and expands the internal storage if more
     * space is needed.
     * 
     * The internal container array can contain a allocated maximum number of elements.
     * Since we need the flexibility to add an indefinite number of elements, the method
     * will expand the internal storage when it detects that we have reached the end
     * of available locations in the current array.
     * 
     * Observe that we are increasing the size member with a postfix operator.
     */
    @Override
    public void add(String newElem) {
        if (size == array.length)
            array = Arrays.copyOf(array, array.length * 2);
        array[size++] = newElem;
    }

    /**
     * Removes the element at a given location of the list.
     * 
     * The methods copies all the elements after the specified index to the same array,
     * over-writting the element being removed. After that we erase the last element
     * of the array by setting it to null, and decrease the size with a postfix operator. 
     * 
     *  @param index Index of the element being removed.
     *  
     */
    @Override
    public void remove(int index) {
        checkIndex(index);
        System.arraycopy(array, index + 1, array, index, size - 1 - index);
        array[--size] = null;
    }

    @Override
    public String get(int index) {
        checkIndex(index);
        return array[index];
    }

    @Override
    public void set(int index, String elem) {
        checkIndex(index);
        array[index] = elem;
    }

    @Override
    public StringIterator iterator() {
        return new StringArrayListIterator(this);
    }

    @Override
    public String toString() {
        StringBuilder b = new StringBuilder("[");
        for (int i = 0; i < size; ++i) {
            if (i > 0)
                b.append(",");
            b.append(array[i]);
        }
        b.append("]");
        return b.toString();
    }

    private void checkIndex(int index) {
        if (! (0 <= index && index < size))
            throw new IndexOutOfBoundsException("invalid index: "+ index +" (must be in [0;"+ size +"[)");
    }
}

/**
 * Implementation of StringIterator for the StringArrayList.
 * 
 * @see StringArrayList#iterator()
 *
 */
class StringArrayListIterator implements StringIterator {
    private final StringArrayList list;
    private int pos = 0;

    StringArrayListIterator(StringArrayList list) {
        this.list = list;
    }

    public boolean hasNext() {
        return pos < list.size();
    }

    public String next() {
        return list.get(pos++);
    }
}