package ch.epfl.collections;

/**
 * Double-linked circular implementation of the StringList interface.
 * 
 * The class manages and keeps track of SLLNode elements which contain the
 * list data.
 * 
 * @see SLLNode
 * @see SLLIterator
 * @see StringLinkedListTest
 *
 */
public class StringLinkedList implements StringList {
    /**
     * Main reference to the list of SLLNode elements. Tail of the list
     * can always be found trough the head.prev 
     */
    private SLLNode head = null;
    private int size = 0;

    @Override
    public boolean isEmpty() {
        return size == 0;
    }

    @Override
    public int size() {
        return size;
    }

    /**
     * Adds new element to the list.
     * 
     * Creates a new SLLNode and updates the references of existing list elements
     * to position the new item at the tail position of the list.
     */
    @Override
    public void add(String newElem) {
        final SLLNode newNode = new SLLNode(newElem);
        if (size == 0) {	// in case of the empty list
            assert head == null;
            head = newNode;
            newNode.next = newNode.prev = newNode;
        } else {			// if the list is not empty, find the tail and update references
            assert head != null;
            SLLNode tail = head.prev;
            tail.next = newNode;
            newNode.prev = tail;
            newNode.next = head;
            head.prev = newNode;
        }
        size += 1;
    }

    /**
     * Removes the item from the list.
     * 
     * When removing item N, the method updates the .next reference of item N-1
     * and the .prev reference of the item N+1.
     */
    @Override
    public void remove(int index) {
        SLLNode nodeToRemove = nodeAtIndex(index);
        nodeToRemove.prev.next = nodeToRemove.next;
        nodeToRemove.next.prev = nodeToRemove.prev;		
        --size;
        if (nodeToRemove == head)
            head = (size == 0 ? null : head.next);
    }

    @Override
    public String get(int index) {
        return nodeAtIndex(index).value;
    }

    @Override
    public void set(int index, String elem) {
        nodeAtIndex(index).value = elem;
    }

    @Override
    public StringIterator iterator() {
        return new SLLIterator(head, size);
    }

    @Override
    public String toString() {
        StringBuilder b = new StringBuilder("[");
        StringIterator it = iterator();
        while (it.hasNext()) {
            b.append(it.next());
            if (it.hasNext())
                b.append(",");
        }
        b.append("]");
        return b.toString();
    }

    /**
     * Internal method for accessing items by iterating N times from the head node.
     * 
     * @param index Index of the element to be returned
     * @return SLLNode at the requested position
     * @throws IndexOutOfBoundsException if index is greater then the number of elements in
     * 		   the list.
     */
    private SLLNode nodeAtIndex(int index) {
        if (index >= size)
            throw new IndexOutOfBoundsException("invalid index: "+ index +" (must be in [0;"+ (size() - 1) + "])");

        SLLNode n = head;
        for (int i = 0; i < index; ++i)
            n = n.next;
        return n;
    }
}

/**
 * Double-linked list element. Used in StringLinkedList.
 *
 * @see StringLinkedList
 */
class SLLNode {
    String value;
    /**
     * References to the previous and next elements of the list. 
     */
    SLLNode prev = null, next = null;

    public SLLNode(String value) {
        this.value = value;
    }
}

/**
 * Iterator for the double-linked implementation of the StringList.
 * 
 * @see StringLinkedList
 */
class SLLIterator implements StringIterator {
    private SLLNode nextNode;
    private int nbLeft;

    public SLLIterator(SLLNode head, int size) {
        nextNode = head;
        nbLeft = size;
    }

    @Override
    public boolean hasNext() {
        return nbLeft > 0;
    }

    @Override
    public String next() {
        String elem = nextNode.value;
        nextNode = nextNode.next;
        --nbLeft;
        return elem;
    }
}
